//https://www.arduino.cc/reference/en/libraries/wifi/
//https://docs.espressif.com/projects/esp-idf/en/latest/esp32/api-reference/system/freertos.html
//https://github.com/ayushsharma82/WebSerial


#include <WiFi.h>
#include <Arduino.h>
#include <ESPAsyncWebServer.h>
#include <WebSerial.h>
#include <AsyncTCP.h>

#define WINDOW_SIZE 5
#define RXD2 16
#define TXD2 17

//7segment not working bcs LCD wrongs pins on PCB
int Data =  4;  
int Latch = 15; 
int Clock= 12; 

//Wifi andmestik
const char* ssid = "Internetu";
const char* password =  "Auravesi";
AsyncWebServer server(80);

int proovi_number = 0; //
static int running_process_number = 1;
static bool conductivity = false;
static int temperature = 0;

//moving avarage filtri jaoks TEMPERATUUR
int INDEX = 0;
int SUM = 0;
int READINGS[WINDOW_SIZE];
int AVERAGED = 0;

//conductivity avarage
int INDEX_conductivity = 0;
int SUM_conductivity = 0;
int READINGS_conductivity[WINDOW_SIZE];
int AVERAGED_conductivity = 0;


TaskHandle_t Task1;//Multithread
TaskHandle_t Task2;
TaskHandle_t Task3;
//TaskHandle_t Task4;
static SemaphoreHandle_t mutex_temperature;
static SemaphoreHandle_t mutex_sendDateToWebserver;


void setup() {
  Serial.begin(115200);
  Serial2.begin(115200, SERIAL_8N1, 16, 17);
  pinMode(Data,OUTPUT);
  pinMode(Latch,OUTPUT);
  pinMode(Clock,OUTPUT);

  WiFi.begin(ssid, password); //Connecto to wifi
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    //Serial.println("Proovin ühendust saada... ");
    //Serial.print(proovi_number); 
    proovi_number +=1;
    if(proovi_number> 20){
      //Serial.println("NOT CONNECTED TO WIFI");
      break;
      }
    }
  if(WiFi.status() == WL_CONNECTED){
    Serial.print("WiFi IP: ");
    Serial.println(WiFi.localIP());
    WebSerial.begin(&server);
    server.begin();
 }
 mutex_temperature = xSemaphoreCreateMutex();
 mutex_sendDateToWebserver = xSemaphoreCreateMutex();

 //start multi thread
  xTaskCreatePinnedToCore(
                    WiFiTask,   /* function. */
                    "Send data via wifi",     /* name  */
                    10000,       /* Stack size of task */
                    NULL,        /* parameter of the task */
                    1,           /* priority of the task */
                    &Task1,      /* Task handle to keep track of created task */
                    0);          /* pin task to core 0 */                  
  delay(200); 


  xTaskCreatePinnedToCore(
                    Task2code,   
                    "MainAlgorithm",    
                    10000,      
                    NULL,       
                    1,          
                    &Task2,      /* Task handle to keep track of created task */
                    1);         
  delay(200); 

  xTaskCreatePinnedToCore(
                    PassiveChecking,   /* function. */
                    "Control data sensors do not indacate that we are at the wrong task",     /* name  */
                    10000,       /* Stack size of task */
                    NULL,        /* parameter of the task */
                    1,           /* priority of the task */
                    &Task3,      /* Task handle to keep track of created task */
                    0);          /* pin task to core 0 */                  
  delay(200); 
}
    
  
  
void WiFiTask( void * pvParameters ){
  for(;;){
    if(WiFi.status() != WL_CONNECTED){ //try to reconnecto to wifi
      int reconnect = 0;
      WiFi.begin(ssid, password);
      while (WiFi.status() != WL_CONNECTED) {
        reconnect++;
        vTaskDelay(500 / portTICK_PERIOD_MS);
        if(reconnect == 2){
          break;
        }
      }
    }
    
    if(WiFi.status() == WL_CONNECTED){//send data to webserial
      WebSerial.print("Is there conductivity ? ");
      WebSerial.println(conductivity); //Toimub samas tuumas pole vaja mutexiga kaitsta
      WebSerial.print("Temperature : ");
      WebSerial.println(temperature);
      vTaskDelay(2000 / portTICK_PERIOD_MS);

  }
}
}
  void Task2code( void * pvParameters ){ //main algortim mida jooksutaktase
  for(;;){
    vTaskDelay(500 / portTICK_PERIOD_MS); //algortimika ei võta arvesse kui pannakse reaktroisse taasdestileerimisele minevat hapet
    switch (running_process_number) {
      
      case 1://Emptying
       sendtoWebserver("Case 1");
       sendDataToSwitchingBoard("%01,02,03%04,12,10%ON"); // closed;opened;machines_state
       vTaskDelay(10000 / portTICK_PERIOD_MS);// Current value is for testing change to 300000 == 5 min for real system
       sendtoWebserver("Case 1 finished");
       running_process_number = 2;
       break;

      case 2: //Washing
        sendtoWebserver("Case 2");
        sendDataToSwitchingBoard("%03,09%04,05,06,12%ON");
        while(true){
        vTaskDelay(10000 / portTICK_PERIOD_MS);     //change delay to 1000 for testing its 10K
          if(conductivity == conductivity){
            break;
          }
        }
        sendtoWebserver("Case 2 finished");
        running_process_number = 3;
        break;
        
      case 3: //Air intake
        sendtoWebserver("Case 3");
        sendDataToSwitchingBoard("%01,02,03%10%ON");
        vTaskDelay(10000 / portTICK_PERIOD_MS); // Current value is for testing change to 180000 == 3 min for real system
        sendtoWebserver("Case 3 finished");
        running_process_number =4;
        break;
        
       case 4: //Washing end
        sendtoWebserver("Case 4");
        sendDataToSwitchingBoard("%09%12,13,04%ON");
        vTaskDelay(10000 / portTICK_PERIOD_MS); // Current value is for testing change to 300000 == 5 min for real system
        sendtoWebserver("Case 4 finished");
        running_process_number =5;
        break;

       case 5: //Loading the acid
        sendtoWebserver("Case 5");
        sendDataToSwitchingBoard("%12,04,05,06%10,09%ON");
        while(true){
          vTaskDelay(10000 / portTICK_PERIOD_MS);     //change delay to 1000 for testing its 10K
          if(conductivity == conductivity){
            break;
          }
        }
        sendtoWebserver("Case 5 finished");
        running_process_number =6;
        break;


       case 6: // Wait for destilation
        sendtoWebserver("Case 6");
        sendDataToSwitchingBoard("%01,02,03%0%ON");
        while(true){
          vTaskDelay(10000 / portTICK_PERIOD_MS);     //change delay to 1000 for testing its 10K
          if(temperature >= temperature){ //X temp, TBD Urmas and pressure X
            break;
          }
        }
        sendtoWebserver("Case 6 finished");
        running_process_number =7;
        break;
       
       case 7: //Ending of first destilate
        sendtoWebserver("Case 7");
        sendDataToSwitchingBoard("%10%4,11%ON");
        while(true){
          vTaskDelay(10000 / portTICK_PERIOD_MS);     //change delay to 1000 for testing its 10K
          if(temperature >= temperature){ //X temp, TBD Urmas and pressure X
            break;
          }
        }
        sendtoWebserver("Case 7 finished");
        running_process_number =8;
        break;

      case 8: //II fraction
        sendtoWebserver("Case 8");
        sendDataToSwitchingBoard("%04,05,06,11%10%ON");
        while(true){
          vTaskDelay(10000 / portTICK_PERIOD_MS);     //change delay to 1000 for testing its 10K
          if(temperature >= temperature){ //X temp, TBD Urmas and pressure X
            break;
          }
        }
        sendtoWebserver("Case 8 finished");
        running_process_number =9;
        break;
        
       case 9: //II fraction
        sendtoWebserver("Case 9");
        sendDataToSwitchingBoard("%10%11,5%ON");
        vTaskDelay(10000 / portTICK_PERIOD_MS); // 5min == 300000 CHANGE
        sendtoWebserver("Case 9 finished");
        running_process_number =10;
        break;

       case 10: //Main run
        sendtoWebserver("Case 10");
        sendDataToSwitchingBoard("%11,04,05,06%10%ON");
         while(true){
          vTaskDelay(10000 / portTICK_PERIOD_MS);     //change delay to 1000 for testing its 10K
          if(temperature >= temperature){ //X temp, TBD Urmas and pressure X
            break;
          }
        }
        sendtoWebserver("Case 10 finished");
        running_process_number =11;
        break;

       case 11: //Ending final run
        sendtoWebserver("Case 11");
        sendDataToSwitchingBoard("%01,02,03,09,11%04,12%OFF");
        while(true){
          vTaskDelay(10000 / portTICK_PERIOD_MS);     //change delay to 1000 for testing its 10K
          if(temperature >= temperature){ //X temp, TBD Urmas and pressure X
            break;
          }
        }
        sendtoWebserver("Case 11 finished");
        running_process_number =1;
        break;

      default:
        break;
  }   
 }
}

void PassiveChecking( void * pvParameters ){
  for(;;){
    vTaskDelay(100 / portTICK_PERIOD_MS);
    getConductivityNTemperatureData();
    vTaskDelay(100 / portTICK_PERIOD_MS);
    segment_display_data(temperature);
  }
}
void segment_display_data(int informationtodisplay){
  
  int tens = informationtodisplay / 10;      // tens 
  int ones = informationtodisplay % 10;      // ones 

  digitalWrite(Latch, LOW);
  shiftOut(Data, Clock, MSBFIRST, tens);
  digitalWrite(Latch, HIGH);

  digitalWrite(Latch, LOW);
  shiftOut(Data, Clock, MSBFIRST, ones);
  digitalWrite(Latch, HIGH);
  
} 

void sendDataToSwitchingBoard(char* to_be_sent_data){
  Serial2.write(to_be_sent_data); 
}

void sendtoWebserver(String informatsioon){
  if( xSemaphoreTake(mutex_sendDateToWebserver,10/portTICK_PERIOD_MS ) == pdTRUE ){
    if(WiFi.status() == WL_CONNECTED){
      WebSerial.println(informatsioon);
  }
  xSemaphoreGive(mutex_sendDateToWebserver);
  }
}


void getConductivityNTemperatureData(){
  if( xSemaphoreTake(mutex_temperature,10/portTICK_PERIOD_MS ) == pdTRUE ){
    int bytesSent = Serial.write("15 poll\n"); //informatsiooni kättesaamine
    if (Serial.available() > 0) {
    String tulemus = Serial.readString();
    String temp = tulemus.substring(tulemus.indexOf("TEMP: ") + 6, tulemus.lastIndexOf("'C") - 5); //TEMP: 25.000000'C
    String conduct = tulemus.substring(tulemus.indexOf("conductivity: ") + 24, tulemus.lastIndexOf("S/cm")); //näide väljundist 4.190333e-05   oli 14
//    Serial.println(temp);
    int temp_muutuja = temp.toInt();
    int conduct_muutuja = conduct.toInt();

    //moving avarage temperatuur
    SUM = SUM - READINGS[INDEX];         
    READINGS[INDEX] = temp_muutuja;          
    SUM = SUM + temp_muutuja;                
    INDEX = (INDEX+1) % WINDOW_SIZE;
    temperature = SUM / WINDOW_SIZE;
    


    SUM_conductivity = SUM_conductivity - READINGS_conductivity[INDEX_conductivity];
    if(conduct_muutuja < 6){ // when the power of conductivity 10**x , x<6 theres is conduction
      READINGS_conductivity[INDEX_conductivity] = 1;
      SUM_conductivity = SUM_conductivity + 1;                 

    }
    else{
      READINGS_conductivity[INDEX_conductivity] = 0;
      SUM_conductivity = SUM_conductivity + 0;                 

    }
    INDEX_conductivity = (INDEX_conductivity+1) % WINDOW_SIZE;
    AVERAGED_conductivity = SUM_conductivity / WINDOW_SIZE;  
    if(AVERAGED_conductivity == 1){
      conductivity = true;
    }
    else{
      conductivity = false;
    }
    }
    xSemaphoreGive(mutex_temperature);
 }   
}

void loop() {
  // put your main code here, to run repeatedly:
}
